@@ -114,7 +114,7 @@ main (void)
    */
   sd = tcp_connect ();
 
-  gnutls_transport_set_ptr (session, (gnutls_transport_ptr_t) sd);
+  gnutls_transport_set_int (session, sd);
 
   /* Perform the TLS handshake
    */
@@ -128,7 +128,11 @@ main (void)
     }
   else
     {
-      printf ("- Handshake was completed\n");
+      char* desc;
+      
+      desc = gnutls_session_get_desc(session);
+      printf ("- Session info: %s\n", desc);
+      gnutls_free(desc);
     }
 
   gnutls_record_send (session, MSG, strlen (MSG));
