@@ -7,7 +7,7 @@
  *
  * The GnuTLS is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public License
- * as published by the Free Software Foundation; either version 3 of
+ * as published by the Free Software Foundation; either version 2.1 of
  * the License, or (at your option) any later version.
  *
  * This library is distributed in the hope that it will be useful, but
@@ -28,7 +28,6 @@
 #include <gnutls_sig.h>
 #include <gnutls_str.h>
 #include <gnutls_datum.h>
-#include <hash.h>
 #include "x509_int.h"
 #include <common.h>
 #include <base64.h>
@@ -59,9 +58,8 @@ int store_pubkey(const char* db_name, const char* host,
                  const char* service, time_t expiration, const gnutls_datum_t* pubkey);
 
 static int find_config_file(char* file, size_t max_size);
-#define MAX_FILENAME 512
 
-void *_gnutls_file_mutex;
+extern void *_gnutls_file_mutex;
 
 struct gnutls_tdb_int default_tdb = {
   store_pubkey,
@@ -170,7 +168,7 @@ uint8_t hphash[MAX_HASH_SIZE*2+1];
   if (p == NULL)
     return gnutls_assert_val(GNUTLS_E_PARSING_ERROR);
     
-  if (p[0] != '*' && strcmp(p, host) != 0)
+  if (p[0] != '*' && host != NULL && strcmp(p, host) != 0)
     return gnutls_assert_val(GNUTLS_E_PARSING_ERROR);
 
   /* read service */
@@ -178,7 +176,7 @@ uint8_t hphash[MAX_HASH_SIZE*2+1];
   if (p == NULL)
     return gnutls_assert_val(GNUTLS_E_PARSING_ERROR);
     
-  if (p[0] != '*' && strcmp(p, service) != 0)
+  if (p[0] != '*' && service != NULL && strcmp(p, service) != 0)
     return gnutls_assert_val(GNUTLS_E_PARSING_ERROR);
 
   /* read expiration */
@@ -259,7 +257,7 @@ time_t expiration;
   if (p == NULL)
     return gnutls_assert_val(GNUTLS_E_PARSING_ERROR);
 
-  if (p[0] != '*' && strcmp(p, host) != 0)
+  if (p[0] != '*' && host != NULL && strcmp(p, host) != 0)
     return gnutls_assert_val(GNUTLS_E_PARSING_ERROR);
 
   /* read service */
@@ -267,7 +265,7 @@ time_t expiration;
   if (p == NULL)
     return gnutls_assert_val(GNUTLS_E_PARSING_ERROR);
     
-  if (p[0] != '*' && strcmp(p, service) != 0)
+  if (p[0] != '*' && service != NULL && strcmp(p, service) != 0)
     return gnutls_assert_val(GNUTLS_E_PARSING_ERROR);
 
   /* read expiration */
@@ -362,7 +360,7 @@ static int raw_pubkey_to_base64(const gnutls_datum_t* raw, gnutls_datum_t * b64)
   char* out;
   
   ret = base64_encode_alloc((void*)raw->data, raw->size, &out);
-  if (ret == 0)
+  if (ret == 0 || out == NULL)
     return gnutls_assert_val(GNUTLS_E_MEMORY_ERROR);
   
   b64->data = (void*)out;
@@ -701,7 +699,7 @@ FILE* fd = NULL;
 int ret;
 char local_file[MAX_FILENAME];
 
-  if (hash_algo == GNUTLS_DIG_MD5 || hash_algo == GNUTLS_DIG_MD2)
+  if (_gnutls_digest_is_secure(hash_algo) == 0)
     return gnutls_assert_val(GNUTLS_E_ILLEGAL_PARAMETER);
     
   if (_gnutls_hash_get_algo_len(hash_algo) != hash->size)
